#!/usr/bin/env node

// src/cli.ts
import cac from "cac";
import { bundleRequire } from "bundle-require";

// src/util.ts
import escalade from "escalade/sync";
import { mkdirSync } from "fs";
import path from "path";
function getOutputFile(file, format) {
  const id = Math.random().toString(36).substring(2, 15);
  const cacheDir = path.join(nearestNodeModules(file), ".cache/codegentool");
  const outputFile = path.join(
    cacheDir,
    file.replace(
      /\.([mc]?[tj]s|[tj]sx)$/,
      `.bundled_${id}.${format === "esm" ? "mjs" : "cjs"}`
    )
  );
  mkdirSync(path.dirname(outputFile), { recursive: true });
  return outputFile;
}
function nearestNodeModules(file) {
  const dir = escalade(path.dirname(file), (dir2, names) => {
    if (names.includes("node_modules")) {
      return dir2;
    }
  });
  if (dir) {
    return path.join(dir, "node_modules");
  }
  return "node_modules";
}

// src/cli.ts
import { watch } from "chokidar";

// src/config.ts
import escalade2 from "escalade/sync";
import fs from "fs";
function readConfig(cwd) {
  let config;
  let configPath;
  const watchPaths = [];
  escalade2(cwd, (dir, names) => {
    watchPaths.push(dir + "/package.json");
    if (names.includes("package.json")) {
      let pkg;
      try {
        pkg = JSON.parse(fs.readFileSync(dir + "/package.json", "utf8"));
      } catch {
      }
      if (pkg?.codegentool && typeof pkg.codegentool === "object") {
        config = pkg.codegentool;
        configPath = dir + "/package.json";
        return dir;
      }
    }
    if (names.includes(".git")) {
      return dir;
    }
  });
  config = {
    generators: ["generators/**/*.{ts,mts,js,mjs}", "!**/node_modules/**"],
    format: "cjs",
    ...config
  };
  return {
    config,
    configPath,
    watchPaths
  };
}

// src/cli.ts
import { dequal } from "dequal";
import fs2 from "fs";
import path2 from "path";
import glob from "fast-glob";
import dedent from "dedent";
import kleur from "kleur";
var program = cac("codegentool");
program.command("").option("--watch, -w", "Enable file watching and automatic reruns").action((options) => start(process.cwd(), options));
program.parse();
async function start(cwd, options) {
  const { config, configPath, watchPaths } = readConfig(cwd);
  if (configPath) {
    process.chdir(path2.dirname(configPath));
  }
  const watcher = watch([...config.generators, ...watchPaths]);
  const generators = /* @__PURE__ */ new Map();
  let initialized = false;
  watcher.on("all", (event, name) => {
    if (name.endsWith("package.json")) {
      if (!initialized)
        return;
      const configResult = readConfig(process.cwd());
      if (!dequal(config, configResult.config) || !dequal(watchPaths, configResult.watchPaths)) {
        watcher.close();
        generators.forEach(async (close) => (await close)());
        start(cwd, options);
      }
    } else if (event == "add" || event == "change") {
      const promise = generators.get(name) ?? Promise.resolve();
      promise.then((close) => close?.()).then(() => {
        generators.set(name, generate(name, config));
      });
    }
  });
  await new Promise((resolve) => {
    watcher.once("ready", () => {
      initialized = true;
      resolve();
    });
  });
  if (options.watch) {
    console.log(`\u2318 Watching for changes...`);
  } else {
    watcher.close();
    generators.forEach(async (close) => (await close)());
  }
}
async function generate(generatorPath, config) {
  const watcher = watch([generatorPath], {
    ignoreInitial: true
  });
  let close = Promise.resolve(() => watcher.close());
  watcher.on(
    "all",
    debounce(() => {
      watcher.close();
      close = generate(generatorPath, config);
    }, 500)
  );
  try {
    console.log(
      `\u25B6 Running %s`,
      kleur.cyan(path2.relative(process.cwd(), generatorPath))
    );
    const {
      mod: { default: generator },
      dependencies
    } = await bundleRequire({
      filepath: generatorPath,
      getOutputFile,
      tsconfig: config.tsconfig,
      format: config.format
    });
    if (typeof generator !== "function") {
      throw new Error(`Generator must default export a function`);
    }
    watcher.add(dependencies);
    let filesGenerated = 0;
    const watchPaths = /* @__PURE__ */ new Set();
    const api = {
      scan: (source, options) => {
        const cwd = path2.resolve(options?.cwd || ".");
        if (Array.isArray(source)) {
          source.forEach((s) => watchPaths.add(path2.resolve(cwd, s)));
        } else {
          watchPaths.add(path2.resolve(cwd, source));
        }
        return glob.sync(source, options);
      },
      read: (file, encoding) => {
        watchPaths.add(file);
        return fs2.readFileSync(file, encoding);
      },
      write: (file, data) => {
        if (typeof data === "string") {
          let current;
          try {
            current = fs2.readFileSync(file, "utf8");
          } catch {
          }
          if (data !== current) {
            fs2.mkdirSync(path2.dirname(file), { recursive: true });
            fs2.writeFileSync(file, data);
            filesGenerated++;
            console.log(
              `\u2714\uFE0F Generated %s`,
              kleur.green(path2.relative(process.cwd(), file))
            );
          }
        } else {
          let current;
          try {
            current = fs2.readFileSync(file);
          } catch {
          }
          if (!current?.equals(data)) {
            fs2.mkdirSync(path2.dirname(file), { recursive: true });
            fs2.writeFileSync(file, data);
            filesGenerated++;
            console.log(
              `\u2714\uFE0F Generated %s`,
              kleur.green(path2.relative(process.cwd(), file))
            );
          }
        }
      },
      dedent
    };
    await generator(api);
    if (!filesGenerated) {
      console.log(
        `\u2714\uFE0F Nothing updated by %s`,
        kleur.cyan(path2.relative(process.cwd(), generatorPath))
      );
    }
    watcher.add([...watchPaths]);
  } catch (error) {
    console.error(`Error in ${path2.resolve(generatorPath)}`);
    console.error(error);
  }
  return async () => (await close)();
}
function debounce(fn, wait) {
  let timeout;
  return function(...args) {
    clearTimeout(timeout);
    timeout = setTimeout(() => fn.apply(this, args), wait);
  };
}
